// Handler for raw blob downloads
//
// Blobs are read via `git cat-file ...` with first querying authentication
// backend about download-access permission for containing repository.
// Replies from authentication backend are cached for 30 seconds to keep
// access-to-blobs latency to minimum.

package main

import (
	"bufio"
	"errors"
	"fmt"
	"io"
	"log"
	"net/http"
	"net/http/httptest"
	"net/url"
	"regexp"
	"strings"
	"sync"
	"time"
)

// Reply from auth backend for "download from repo" authorization request
type AuthReply struct {
	// raw reply from auth backend & preAuthorizeHandler().
	// recorded so we can replay it from auth cache to each client in full
	// if e.g. access is rejected.
	RawReply *httptest.ResponseRecorder

	// decoded auth reply
	authorizationResponse
}

// Entry in authorization reply cache
type AuthCacheEntry struct {
	// FIXME we need to lock the entry only to "correctly" update Nhit on
	// read side, but we can tolerate some looses in Nhit and update it
	// without mutex or atomic. Only -race complains...
	// ( we could use atomic.Value for atomic whole cache entry updates from
	//   refresher without need for locks on readers side, but the need to do
	//   .Nhit++ on readers side ruins that )
	sync.Mutex

	AuthReply

	Tauth int64         // in seconds
	Nhit  int64         // how many times this entry was hit when querying auth cache during the last refresh period.
	ready chan struct{} // closed when entry is ready
}

// Entries are keyed by project + credentials
type AuthCacheKey struct {
	project string
	query   string // e.g. with passing in private_token=...
	header  string // request header url-encoded, e.g. PRIVATE-TOKEN=...
}

// Authorization reply cache
// {} AuthCacheKey -> *AuthCacheEntry
type AuthCache struct {
	u *upstream // for which upstream we cache auth

	mu     sync.RWMutex // guards .cached
	cached map[AuthCacheKey]*AuthCacheEntry
}

func NewAuthCache(u *upstream) *AuthCache {
	return &AuthCache{u: u, cached: make(map[AuthCacheKey]*AuthCacheEntry)}
}

// Verify that download access is ok or not.
// first we try to use the cache; if information is not there -> ask auth backend
// download is ok if AuthReply.RepoPath != ""
func (c *AuthCache) VerifyDownloadAccess(project string, query url.Values, header http.Header) AuthReply {
	// Use only tokens from query/header to minimize cache and avoid
	// creating redundant cache entries because of e.g. unrelated headers.
	q := url.Values{}
	for k, v := range query {
		if strings.HasSuffix(k, "_token") {
			q[k] = v
		}
	}
	h := url.Values{}
	for k, v := range header {
		if strings.HasSuffix(strings.ToUpper(k), "-TOKEN") {
			h[k] = v
		}
	}

	key := AuthCacheKey{project, q.Encode(), h.Encode()}
	return c.verifyDownloadAccess(key)
}

func (c *AuthCache) verifyDownloadAccess(key AuthCacheKey) AuthReply {
	var authReply AuthReply

	// first try to read from cache in parallel with other readers
	c.mu.RLock()
	auth := c.cached[key]
	c.mu.RUnlock()

have_entry:
	// entry in cache - use it
	if auth != nil {
		<-auth.ready // wait untill it is ready

		auth.Lock()
		auth.Nhit++
		//log.Printf("authReply for %v cached ago: %v  (hits: %v)",
		//	key,
		//	time.Since(time.Unix(auth.Tauth, 0)),
		//	auth.Nhit)
		authReply = auth.AuthReply
		auth.Unlock()

	} else {
		// no entry - relock the cache in exclusive mode, create empty entry,
		// and start filling it

		c.mu.Lock()
		// another ex-reader could be trying to create this entry
		// simultaneously with us - recheck
		auth = c.cached[key]
		if auth != nil {
			c.mu.Unlock()
			goto have_entry
		}

		// new not-yet-ready entry
		auth = &AuthCacheEntry{ready: make(chan struct{})}
		c.cached[key] = auth
		c.mu.Unlock()

		// this goroutine becomes responsible for querying auth backend
		auth.AuthReply = c.askAuthBackend(key)
		auth.Tauth = time.Now().Unix()
		auth.Nhit = 0

		authReply = auth.AuthReply

		// broadcast to other goroutines that this entry is ready
		close(auth.ready)

		// launch entry refresher
		go c.refreshEntry(auth, key)
	}

	return authReply
}

// Time period for refreshing / removing unused entires in authCache
const authCacheRefresh = 30 * time.Second

// Goroutine to refresh auth cache entry periodically while it is used.
// if the entry is detected to be not used - remove it from cache and stop refreshing.
func (c *AuthCache) refreshEntry(auth *AuthCacheEntry, key AuthCacheKey) {
	for {
		time.Sleep(authCacheRefresh)

		auth.Lock()
		nhit := auth.Nhit
		auth.Unlock()

		// clear cache entry if it is not used
		//log.Printf("AUTH refresh - %v #hit: %v", key, nhit)
		if nhit == 0 { // not used - we can remove and stop refreshing
			//log.Printf("AUTH - removing %v", key)
			// NOTE it is ok even if someone gets this auth in this time window
			//      and use it for some time
			c.mu.Lock()
			delete(c.cached, key)
			c.mu.Unlock()
			break
		}

		//log.Printf("AUTH - refreshing %v", key)
		authReply := c.askAuthBackend(key)

		auth.Lock()
		auth.AuthReply = authReply
		auth.Tauth = time.Now().Unix()
		auth.Nhit = 0
		auth.Unlock()
	}
}

// Ask auth backend about cache key
func (c *AuthCache) askAuthBackend(key AuthCacheKey) AuthReply {
	// key.header -> url.Values -> http.Header
	hv, err := url.ParseQuery(key.header)
	if err != nil {
		// we prepared key.header ourselves via url-encoding in AuthCache.VerifyDownloadAccess().
		// It must be ok
		panic(err)
	}
	header := make(http.Header)
	for k, v := range hv {
		header[k] = v
	}

	return askAuthBackend(c.u, key.project, key.query, header)
}

// Ask auth backend about whether download is ok for a project.
// Authorization is approved if AuthReply.RepoPath != "" on return
// Raw auth backend response is emitted to AuthReply.RawReply
func askAuthBackend(u *upstream, project, query string, header http.Header) AuthReply {
	authReply := AuthReply{
		RawReply: httptest.NewRecorder(),
	}

	// Request to auth backend to verify whether download is possible.
	// - first option is via asking as `git fetch` would do, but on Rails
	//   side this supports only basic auth, not private token.
	// - that's why we auth backend to authenticate as if it was request to
	//   get repo archive and propagate request query and header.
	// url := project + ".git/info/refs?service=git-upload-pack"
	url := project + "/repository/archive.zip"
	if query != "" {
		url += "?" + query
	}
	reqDownloadAccess, err := http.NewRequest("GET", url, nil)
	if err != nil {
		fail500(authReply.RawReply, fmt.Errorf("GET git-upload-pack: %v", err))
		return authReply
	}
	for k, v := range header {
		reqDownloadAccess.Header[k] = v
	}

	// Prepare everything and go through preAuthorizeHandler() that will send
	// request to auth backend and analyze/parse the reply into r.authorizationResponse.
	// it also logs/emits output in case of errors - we do not have to do it here
	r := &gitRequest{
		Request: reqDownloadAccess,
		u:       u,
	}

	preAuthorizeHandler(
		func(w http.ResponseWriter, r *gitRequest) {
			// if we ever get to this point - auth handler approved
			// access and thus it is ok to download
			authReply.authorizationResponse = r.authorizationResponse
		}, "")(authReply.RawReply, r)

	return authReply
}

// HTTP handler for `.../raw/<ref>/path`
var rawRe = regexp.MustCompile(`/raw/`)

func handleGetBlobRaw(w http.ResponseWriter, r *gitRequest) {
	// Extract project & refpath
	// <project>/raw/branch/file -> <project>, branch/file
	u := r.Request.URL
	rawLoc := rawRe.FindStringIndex(u.Path)
	if rawLoc == nil {
		fail500(w, errors.New("extract project name"))
		return
	}
	project := u.Path[:rawLoc[0]]
	refpath := u.Path[rawLoc[1]:]

	// Query download access auth for this project
	authReply := r.u.authCache.VerifyDownloadAccess(project, u.Query(), r.Request.Header)
	if authReply.RepoPath == "" {
		// access denied - copy auth reply to client in full -
		// there are HTTP code and other headers / body relevant for
		// about why access was denied.
		for k, v := range authReply.RawReply.HeaderMap {
			w.Header()[k] = v
		}
		w.WriteHeader(authReply.RawReply.Code)
		// NOTE do not consume authReply.RawReply.Body with io.Copy() -
		// this way it will be read one time only and next reads will be empty.
		_, err := w.Write(authReply.RawReply.Body.Bytes())
		if err != nil {
			logError(fmt.Errorf("writing authReply.RawReply.Body: %v", err))
		}
		return
	}

	// Access granted - we can emit the blob
	emitBlob(w, authReply.RepoPath, refpath)
}

// Emit content of blob located at <ref>/path (jointly denoted as 'refpath') to output
func emitBlob(w http.ResponseWriter, repopath string, refpath string) {
	// Communicate with `git cat-file --batch` trying refs from longest
	// to shortest prefix in refpath. This way we find longest-match for
	// ref and get blob sha1 and content in the end.
	queryCmd := gitCommand("", "git", "--git-dir="+repopath, "cat-file", "--batch")
	queryStdin, err := queryCmd.StdinPipe()
	if err != nil {
		fail500(w, fmt.Errorf("git cat-file --batch; stdin: %v", err))
		return
	}
	defer queryStdin.Close()
	queryStdout, err := queryCmd.StdoutPipe()
	if err != nil {
		fail500(w, fmt.Errorf("git cat-file --batch; stdout: %v", err))
		return
	}
	defer queryStdout.Close()
	queryReader := bufio.NewReader(queryStdout)

	err = queryCmd.Start()
	if err != nil {
		fail500(w, fmt.Errorf("git cat-file --batch; start: %v", err))
		return
	}
	defer cleanUpProcessGroup(queryCmd)

	// refpath components as vector
	refpathv := strings.Split(refpath, "/")

	// scan from right to left and try to change '/' -> ':' and see if it
	// creates a correct git object name. If it does - we read object
	// content which follows.
	var sha1, type_ string
	var size int64
	for i := len(refpathv); i > 0; i-- {
		ref := strings.Join(refpathv[:i], "/")
		path := strings.Join(refpathv[i:], "/")
		_, err := fmt.Fprintf(queryStdin, "%s:%s\n", ref, path)
		if err != nil {
			fail500(w, fmt.Errorf("git cat-file --batch; write: %v", err))
			return
		}

		reply, err := queryReader.ReadString('\n')
		if err != nil {
			fail500(w, fmt.Errorf("git cat-file --batch; read: %v", err))
			return
		}

		// <object> SP missing LF
		if strings.HasSuffix(reply, " missing\n") {
			continue
		}

		// <sha1> SP <type> SP <size> LF
		_, err = fmt.Sscanf(reply, "%s %s %d\n", &sha1, &type_, &size)
		if err != nil {
			fail500(w, fmt.Errorf("git cat-file --batch; reply parse: %v", err))
			return
		}

		if type_ != "blob" {
			log.Printf("git cat-file --batch-check; %v is not blob (is %v)", sha1, type_)
			sha1 = "" // so it will return 404
		}

		// git object found
		break
	}

	// Blob not found -> 404
	if sha1 == "" {
		http.Error(w, fmt.Sprintf("Blob for %v not found", refpath), http.StatusNotFound)
		return
	}

	// Blob found - start writing response
	w.Header().Set("Content-Disposition", "inline")
	w.Header().Set("Content-Transfer-Encoding", "binary")
	w.Header().Set("Content-Length", fmt.Sprintf("%d", size))
	w.Header().Set("X-Content-Type-Options", "nosniff")
	// net/http sniffs stream and automatically detects and sets
	// Content-Type header. We do not have to do it ourselves.

	w.Header().Set("Cache-Control", "private") // Rails sets this for IE compatibility
	w.Header().Set("ETag", fmt.Sprintf(`"%s"`, sha1))

	w.WriteHeader(http.StatusOK) // Don't bother with HTTP 500 from this point on, just return
	// XXX better use queryStdout instead of queryReader, but we could be
	// holding some tail bytes in queryReader after chat phase
	_, err = io.CopyN(w, queryReader, size)
	if err != nil {
		logError(fmt.Errorf("io.CopyN: %v", err))
		return
	}

	// close git stdin explicitly, so it can exit cleanly
	err = queryStdin.Close()
	if err != nil {
		logError(fmt.Errorf("queryStdin.Close: %v", err))
		return
	}

	err = queryCmd.Wait()
	if err != nil {
		logError(fmt.Errorf("wait: %v", err))
		return
	}
}
